/*
 * ijsio.cpp
 * Authors: Soren Johnson and Leonid Reyzin (reyzin@cs.bu.edu)
 *
 * The algorithms in this file provide the command-line parsing and
 * input-output routines for ijs.cpp.
 *
 * See ijs.txt for more information.  This code and explanatory notes
 * are hosted at http://www.cs.bu.edu/~reyzin/code/fuzzy.html
 *
 * Uses Victor Shoup's NTL, see http://www.shoup.net
 *
 */

#include "ijs.h"
#include "fstream"

static char* defaultSketchfile = "sketch.out";
static char* defaultRecfile = "rec.out";
static char* defaultModfile = "mod.out";



/*
 * Generates a prime of bit length primLen and writes it out to
 * modfilename
 * In case of error, complains to the user via cerr and exits with code -1
 * DOES NOT CHECK FOR ALL POSSIBLE FILE WRITING PROBLEMS
 */
void ProducePrime (const char * modfilename, long primLen) {
  ZZ mod;
  GenPrime (mod, primLen);
  ofstream modfile (modfilename);
  if (!modfile) {
    cerr << "Error: could not open file "<<modfilename<<" for writing!\n";
    exit(-1);
  }
  modfile<<mod;
  modfile.close();
}

/*
 * Prints out the help message to the user via standard output (cout)
 */
void PrintHelp(const char * programName) {

  cout<<"Usage:  "<<programName<<" [-h] [-s infile (outfile)] [-r recfile sketchfile (outfile)]\n"
      <<"        [-p bits (outfile)]\n\n"
      <<"\t-h prints this help text.\n"
      <<"\t-s creates a secure sketch of infile and writes it to outfile, if\n"
      <<"\t   specified, '"<<defaultSketchfile<<"' otherwise.\n"
      <<"\t-r reconciles the set in recfile with the set sketched in sketchfile\n"
      <<"\t   and writes it to outfile, if specified, '"<<defaultRecfile<<"' otherwise.\n"
      <<"\t-p creates a prime modulus with a given number of bits and writes it\n"
      <<"\t   to outfile, if specified, '"<<defaultModfile<<"' otherwise.\n\n"
      <<"Files that represent a set w = {w_1, w_2, ... w_s} of n elements modulo a\n"
      <<"prime number n with maximum number of differences t should be in the form\n"
      <<"t=value n=value [w_1 w_2 ... w_n].\n\n"  
      <<"The -p command is just a tool for creating a large prime modulus n and\n"
      <<"does not affect the execution of the sketch or reconcile commands.\n\n"
      <<"Implements the Improved Juels-Sudan secure sketch from \"Fuzzy Extractors:\n"
      <<"How to Generate Strong Keys from Biometrics and Other Noisy Data\" by\n"
      <<"Dodis, Ostrovsky, Reyzin, and Smith\n"
      <<"(SIAM Journal on Computing, 38(1):97-139, 2008).\n" 
      <<"See ijs.txt and http://www.cs.bu.edu/~reyzin/code/fuzzy.html for more details.\n";
}



/*
 * Reads a set contains in the file infilename; returns field size n, the error tolerance t,
 * and the set itself via its first three arguments.
 * Initializes NTL's ZZ_p to work with n as the modulus
 * The file should begin with t=value and n=value (no spaces around "=") and then have a
 * vector of integers modulo n
 * In case of error, complains to the user via cerr and exits with code -1
 * DOES NOT CHECK FOR ALL POSSIBLE INPUT FORMATTING ERRORS OR FILE READING/WRITING PROBLEMS
 */
void ReadSet (ZZ & n, int & t, vec_ZZ_p & set, const char * infilename)
{
  ifstream infile(infilename);
  if (!infile) {
    cerr << "Error: could not open file "<<infilename<<" for reading!\n";
    exit(-1);
  }
  

  // Read in 2 parameters: the modulus n and the error tolerance t
  for (int count = 0; count < 2; count++) {
    char c1 = '\0', c2 = '\0';
    while (c2 != '=' && !infile.eof())
      infile >> c1 >> c2;
    switch (c1) {
    case 't': 
      infile >> t;  break;
    case 'n': 
      infile >> n; break;
    default: 
      cerr << "Error: bad input format!  Use the -h option for format help.\n"; infile.close(); exit(-1);
    }
  }
  
  
  if (infile.eof()) {
    cerr << "Error: bad input format!  Use the -h option for format help.\n";
    infile.close();
    exit(-1);
  }
  
  ZZ_p::init(n);   // Initiliaze ZZ_p to work with n as the modulus

  infile>>set;     // Read in the set itself

  infile.close();
  if (t>set.length()) {
    cerr<<"Error: t can be at most the length of the input set,\n  but you specified t="<<t<<" for an input set of length "<<set.length()<<".\n";
    exit(-1);
  }

  if (t<0 || t%2==1) {
   cerr<<"Error: t must be nonnegative and even, but you specified t="<<t<<".\n";
   exit(-1);
  }
}

/*
 * Reads in a set (together with parameter n and t) contained in infilename, produces a sketch,
 * and writes it to outfilename.
 * In case of error, complains to the user via cerr and exits with code -1
 * DOES NOT CHECK FOR ALL POSSIBLE INPUT FORMATTING ERRORS OR FILE READING/WRITING PROBLEMS
 */
void ProduceSketch (const char* infilename, const char* outfilename)
{
  // Read in the set and the initialize ZZ_p
  ZZ n; // the prime field size
  int t; // the error tolerance: we tolerate symmetric difference up to t between two sets
  vec_ZZ_p set; // the set whose sketch we are produce
  ReadSet (n, t, set, infilename);

  vec_ZZ_p sketch;
  IJS_Sketch(sketch, set, t); // produce the actual sketch -- this is where the math happens

  // Write out the sketch to the file
  ofstream outfile (outfilename);
  if (!outfile) {
    cerr << "Error: could not open file "<<outfilename<<" for writing";
    exit (-1);
  }
  outfile << sketch;
  outfile.close();
}




/*
 * Recovers the original set given a different set (with at most t differences)
 * contained in receivedFilename and a sketch of length t contained in sketchFilename.
 * Writes the result to outFilename
 * More formally, this function finds, if it exists,
 * the unique set of the same size as contained in receivedFilename
 * that has the following two properties:
 * (a) Its symmetric difference with the set in receivedFilename is at most t
 * (b) It has the given sketch
 * If such a set does not exist, or if anything else goes wrong,
 * complains to the user via cerr and exits with code -1
 *
 * DOES NOT CHECK FOR ALL POSSIBLE INPUT FORMATTING ERRORS OR FILE READING/WRITING PROBLEMS
 */
void RecoverSet (const char* receivedFilename, const char* sketchFilename, const char* outFilename) {

  // Read in the received set and initialize ZZ_p
  ZZ n; // the prime field size
  int t; // the error tolerance: we tolerate symmetric difference up to t between two sets
  vec_ZZ_p receivedSet;
  ReadSet (n, t, receivedSet, receivedFilename);

  // Read in the sketch
  vec_ZZ_p sketch;
  ifstream sketchfile(sketchFilename);
  if (!sketchfile) {
    cerr << "Error: Could not open file "<<sketchFilename<<" for reading!\n";
    exit(-1);
  }
  sketchfile >> sketch;
  sketchfile.close();

  if (sketch.length() != t) {
    cerr<<"Error: sketch is of an incorrect length "<<sketch.length()<<" instead of "<< t <<endl;
    exit(-1);
  }

  vec_ZZ_p recoveredSet;
  if(!IJS_Recover(recoveredSet, receivedSet, sketch)) {
    cerr<<"Error: number of differences exceeds "<<t<<", the length of the sketch.\n  Recovery not possible.\n";
    exit(-1);
  }

  // Output the recovered set
  ofstream outfile(outFilename);
  if (!outfile) {
    cerr << "Error: could not open file "<<outFilename<<" for writing!\n";
    exit(-1);
  }
  outfile<<"t="<< t << endl <<"n="<<ZZ_p::modulus() << endl  << recoveredSet;
  outfile.close();
}


/*
 * main function reads and parses command-line input and
 * calls the appropriate portion of the program.
 * Complains to the user via cerr if something is wrong
 * Returns -1 on error, 0 if everything goes through
 * DOES NOT CHECK FOR ALL POSSIBLE INPUT FORMATTING ERRORS
 * OR FILE READING/WRITING PROBLEMS
 */
int main (int argc, char** argv) {

  if (argc < 2) {
    cerr<<"Error: No input given.  Use -h for help.\n";
    return -1;
  }

  // first command line parameter should begin with a dash
  if (argv[1][0] != '-') {
    cerr<<"Error: Invalid input.  Use -h for help.\n";
    return -1;
  }

  char c = argv[1][1]; // the option chosen by the user

  
  if ((c == 'h') || (c == 'H')) { // * print out Help *
    PrintHelp(argv[0]);
    return 0;
  }


  if ((c == 's') || (c == 'S')) { // * Create a secure sketch *
    // Usage: sketch -s infile (outfile)
    if (argc <= 2)  {
      cerr<<"Error: No file specified as an input set for sketch creation.  Use -h for help.\n";
      return -1;
    }

    // figure out the output file name
    char * sketchfilename;
    if (argc > 3)
      sketchfilename = argv[3];
    else
      sketchfilename = defaultSketchfile;

    ProduceSketch (argv[2], sketchfilename);
    return 0;
  } 

  if ((c == 'r') || (c == 'R')) { // * Reconcile a set with a sketch already produced
      // Usage: sketch -r recfile sketchfile (outfile)

    if (argc<=3) {
      if (argc == 3)
        cerr<<"Error: No file specified containing the secure sketch with which to reconile "
             <<argv[2]<<".  Use -h for help.\n";
      else
        cerr<<"Error: No file specified to reconcile.  Use -h for help.\n";
      return -1;
    }

    // figure out the output file name
    char * recfilename;
    if (argc > 4)
      recfilename = argv[4];
    else 
      recfilename = defaultRecfile;

    RecoverSet (argv[2], argv[3], recfilename);
    return 0;
  }

  if ((c == 'p') || (c == 'P')) { // * generate a prime modulus
    // Usage: sketch -p bit_length (outfile)

    long primLen;
    if (argc <= 2 || (primLen=atoi(argv[2]))<2) {
      cerr<<"Error: Must specify the number of bits (greater than 1) of the modulus.  Use -h for help.\n";
      return -1;
    }
    char * modfilename;
    // figure out the output file name
    if (argc > 3)
      modfilename = argv[3];
    else
      modfilename = defaultModfile;

    ProducePrime(modfilename, primLen);
    return 0;
  }

  cerr<<"Error: Invalid command.  Use -h for help.\n";
  return -1;
}

